import copy
import itertools

from dune.perftool.loopy.symbolic import substitute
from dune.perftool.pdelab.argument import (name_accumulation_variable,
                                           name_coefficientcontainer,
                                           pymbolic_coefficient,
                                           PDELabAccumulationFunction,
                                           )
from dune.perftool.generation import (backend,
                                      barrier,
                                      built_instruction,
                                      domain,
                                      dump_accumulate_timer,
                                      function_mangler,
                                      generator_factory,
                                      get_counter,
                                      get_global_context_value,
                                      globalarg,
                                      iname,
                                      instruction,
                                      post_include,
                                      kernel_cached,
                                      retrieve_cache_items,
                                      silenced_warning,
                                      temporary_variable,
                                      transform,
                                      )
from dune.perftool.options import get_option
from dune.perftool.loopy.flatten import flatten_index
from dune.perftool.loopy.buffer import (get_buffer_temporary,
                                        initialize_buffer,
                                        switch_base_storage,
                                        )
from dune.perftool.sumfact.quadrature import nest_quadrature_loops
from dune.perftool.pdelab.localoperator import determine_accumulation_space
from dune.perftool.pdelab.restriction import restricted_name
from dune.perftool.pdelab.signatures import assembler_routine_name
from dune.perftool.pdelab.spaces import (name_lfs,
                                         name_lfs_bound,
                                         )
from dune.perftool.pdelab.geometry import (local_dimension,
                                           world_dimension,
                                           )
from dune.perftool.sumfact.amatrix import (AMatrix,
                                           LargeAMatrix,
                                           quadrature_points_per_direction,
                                           basis_functions_per_direction,
                                           construct_amatrix_sequence,
                                           )
from dune.perftool.sumfact.switch import (get_facedir,
                                          get_facemod,
                                          )
from dune.perftool.loopy.symbolic import SumfactKernel
from dune.perftool.ufl.modified_terminals import extract_modified_arguments
from dune.perftool.tools import get_pymbolic_basename
from dune.perftool.error import PerftoolError
from pymbolic.primitives import (Call,
                                 Product,
                                 Subscript,
                                 Variable,
                                 )
from dune.perftool.sumfact.quadrature import quadrature_inames
from dune.perftool.sumfact.vectorization import find_sumfact
from loopy.symbolic import FunctionIdentifier, IdentityMapper

import loopy as lp
import numpy as np
import pymbolic.primitives as prim
from pytools import product


@function_mangler
def horizontal_add_function_mangler(knl, func, arg_dtypes):
    if func == "horizontal_add":
        from dune.perftool.loopy.vcl import get_vcl_type
        vcl = lp.types.NumpyType(get_vcl_type(np.float64, register_size=256))
        return lp.CallMangleInfo("horizontal_add", (lp.types.NumpyType(np.float64),), (vcl,))


@iname
def _sumfact_iname(bound, _type, count):
    name = "sf_{}_{}".format(_type, str(count))
    domain(name, bound)
    return name


def sumfact_iname(bound, _type):
    count = get_counter('_sumfac_iname_{}'.format(_type))
    name = _sumfact_iname(bound, _type, count)
    return name


@kernel_cached
def accum_iname(restriction, bound, i):
    return sumfact_iname(bound, "accum")


def setup_theta(inp, element, restriction, component, index, coeff_func):
    if index is None:
        index = ()
    else:
        index = (index,)
    # Write initial coefficients into buffer
    lfs = name_lfs(element, restriction, component)
    basisiname = sumfact_iname(name_lfs_bound(lfs), "basis")
    container = coeff_func(restriction)
    coeff = pymbolic_coefficient(container, lfs, basisiname)
    assignee = Subscript(Variable(inp), (Variable(basisiname),) + index)
    return instruction(assignee=assignee,
                       expression=coeff,
                       )


def name_test_function_contribution(test):
    count = get_counter('sumfact_contribution_counter')
    grad = "grad_" if test.reference_grad else ""
    return restricted_name("contrib_{}phi_{}".format(grad, str(count)), test.restriction)


@backend(interface="accum_insn", name="sumfact")
def generate_accumulation_instruction(visitor, accterm, measure, subdomain_id):
    pymbolic_expr = visitor(accterm.term)

    if pymbolic_expr == 0:
        return

    dim = world_dimension()

    # If this is a gradient, we find the gradient iname
    additional_inames = frozenset({})
    if accterm.argument.index:
        for i in accterm.argument.index._indices:
            if i not in visitor.dimension_indices:
                from dune.perftool.pdelab.localoperator import grad_iname
                additional_inames = additional_inames.union(frozenset({grad_iname(i, dim)}))

    def emit_sumfact_kernel(i, restriction, insn_dep):
        # Construct the matrix sequence for this sum factorization
        a_matrices = construct_amatrix_sequence(transpose=True,
                                                derivative=i if accterm.argument.index else None,
                                                facedir=get_facedir(accterm.argument.restriction),
                                                facemod=get_facemod(accterm.argument.restriction),
                                                )

        # Get the vectorization info. If this happens during the dry run, we get dummies
        from dune.perftool.sumfact.vectorization import get_vectorization_info
        a_matrices, buf, inp, index, padding = get_vectorization_info(a_matrices, (accterm.argument.restriction, restriction))

        # Initialize a base storage for this buffer and get a temporay pointing to it
        shape = tuple(mat.cols for mat in a_matrices if mat.face is None)
        dim_tags = ",".join(['f'] * local_dimension())
        if index is not None:
            shape = shape + (4,)
            dim_tags = dim_tags + ",c"
            index = (index,)
            vectag = frozenset({"gradvec"})
        else:
            index = ()
            vectag = frozenset()

        temp = initialize_buffer(buf,
                                 base_storage_size=product(max(mat.rows, mat.cols) for mat in a_matrices),
                                 num=2
                                 ).get_temporary(shape=shape,
                                                 dim_tags=dim_tags,
                                                 name=inp,
                                                 )

        # Those input fields, that are padded need to be set to zero in order to do a horizontal_add lateron
        for pad in padding:
            instruction(assignee=prim.Subscript(prim.Variable(temp), tuple(Variable(i) for i in quadrature_inames()) + (pad,)),
                        expression=0,
                        forced_iname_deps=frozenset(quadrature_inames() + visitor.inames),
                        forced_iname_deps_is_final=True,
                        tags=frozenset(["quadvec"])
                        )

        # Replace gradient iname with correct index for assignement
        replace_dict = {}
        for iname in additional_inames:
            replace_dict[Variable(iname)] = i
        expression = substitute(pymbolic_expr, replace_dict)

        # Write timing stuff for jacobian (for alpha methods it is done at the end of stage 1)
        timer_dep = frozenset()
        if get_option("instrumentation_level") >= 4:
            timer_name = assembler_routine_name() + '_kernel' + '_quadratureloop'
            post_include('HP_DECLARE_TIMER({});'.format(timer_name), filetag='operatorfile')
            dump_accumulate_timer(timer_name)
            if(visitor.inames):
                timer_dep = frozenset({instruction(code="HP_TIMER_START({});".format(timer_name),
                                                   within_inames=frozenset(visitor.inames))})

        # Determine dependencies
        from loopy.match import Or, Writes
        from loopy.symbolic import DependencyMapper
        from dune.perftool.tools import get_pymbolic_basename
        deps = Or(tuple(Writes(get_pymbolic_basename(expr)) for expr in DependencyMapper()(expression)))

        # Issue an instruction in the quadrature loop that fills the buffer
        # with the evaluation of the contribution at all quadrature points
        assignee = Subscript(Variable(temp), tuple(Variable(i) for i in quadrature_inames()) + index)
        contrib_dep = instruction(assignee=assignee,
                                  expression=expression,
                                  forced_iname_deps=frozenset(quadrature_inames() + visitor.inames),
                                  forced_iname_deps_is_final=True,
                                  tags=frozenset({"quadvec"}).union(vectag),
                                  depends_on=frozenset({deps}).union(timer_dep)
                                  )

        if insn_dep is None:
            insn_dep = frozenset({contrib_dep})

        if get_option("instrumentation_level") >= 4:
            insn_dep = frozenset({instruction(code="HP_TIMER_STOP({});".format(timer_name),
                                              depends_on=insn_dep,
                                              within_inames=frozenset(visitor.inames))})

        # Add a sum factorization kernel that implements the multiplication
        # with the test function (stage 3)
        pref_pos = i if accterm.argument.index else None
        result, insn_dep = sum_factorization_kernel(a_matrices,
                                                    buf,
                                                    3,
                                                    insn_dep=insn_dep,
                                                    additional_inames=frozenset(visitor.inames),
                                                    preferred_position=pref_pos,
                                                    restriction=(accterm.argument.restriction, restriction),
                                                    )

        inames = tuple(accum_iname((accterm.argument.restriction, restriction), mat.rows, i) for i, mat in enumerate(a_matrices))

        # Collect the lfs and lfs indices for the accumulate call
        test_lfs = determine_accumulation_space(accterm.argument.expr, 0, measure)
        test_lfs.index = flatten_index(tuple(Variable(i) for i in inames),
                                       (basis_functions_per_direction(),) * dim,
                                       order="f"
                                       )

        # In the jacobian case, also determine the space for the ansatz space
        ansatz_lfs = determine_accumulation_space(accterm.term, 1, measure)
        rank = 2 if visitor.inames else 1
        if rank == 2:
            # TODO the next line should get its inames from elsewhere. This is *NOT* robust (but works right now)
            ansatz_lfs.index = flatten_index(tuple(Variable(visitor.inames[i]) for i in range(world_dimension())),
                                             (basis_functions_per_direction(),) * dim,
                                             order="f"
                                             )

        # Construct the expression representing "{r,jac}.accumulate(..)"
        accum = name_accumulation_variable(ansatz_lfs.get_restriction() + test_lfs.get_restriction())

        # Determine the expression to accumulate with. This depends on the vectorization strategy!
        result = prim.Subscript(result, tuple(prim.Variable(i) for i in inames))
        vecinames = ()
        if index:
            iname = accum_iname((accterm.argument.restriction, restriction), 4, "vec")
            vecinames = (iname,)
            transform(lp.tag_inames, [(iname, "vec")])
            from dune.perftool.tools import maybe_wrap_subscript
            result = prim.Call(prim.Variable("horizontal_add"),
                               (maybe_wrap_subscript(result, prim.Variable(iname)),),
                               )

        # In the case of FastDGGridOperator we can write directly into the resiudal/jacobi
        #
        # TODO: At the moment this only works if we do not vectorize
        # (over gradients) because loopy tries to acces a vectorclass
        # variable.
        if get_option('fastdg'):
            ft = get_global_context_value("form_type")
            if ft == 'residual' or ft == 'jacobian_apply':
                accum = accum + ".data()"
                size = basis_functions_per_direction() ** world_dimension()
                globalarg(accum, dtype=np.float64, shape=(size,), managed=False)
                assignee = prim.Subscript(prim.Variable(accum), (test_lfs.index,))
                expression = prim.Sum((assignee, result))
                instruction(assignee=assignee,
                            expression=expression,
                            forced_iname_deps=frozenset(inames),
                            forced_iname_deps_is_final=True,
                            depends_on=insn_dep,
                            )
            else:
                assert ft == 'jacobian'
                accum = accum + ".data()"
                size = basis_functions_per_direction() ** world_dimension()
                globalarg(accum, dtype=np.float64, shape=(size, size), managed=True)
                assignee = prim.Subscript(prim.Variable(accum), (ansatz_lfs.index, test_lfs.index))
                expression = prim.Sum((assignee, result))
                instruction(assignee=assignee,
                            expression=expression,
                            forced_iname_deps=frozenset(inames + visitor.inames),
                            forced_iname_deps_is_final=True,
                            depends_on=insn_dep,
                            )
        # Default: Generate accumulation instructions
        else:
            expr = Call(PDELabAccumulationFunction(accum, rank),
                        (ansatz_lfs.get_args() +
                         test_lfs.get_args() +
                         (result,)
                         )
                        )
            instruction(assignees=(),
                        expression=expr,
                        forced_iname_deps=frozenset(inames + visitor.inames + vecinames),
                        forced_iname_deps_is_final=True,
                        depends_on=insn_dep,
                        )

        # Mark the transformation that moves the quadrature loop inside the trialfunction loops for application
        transform(nest_quadrature_loops, visitor.inames)

        return insn_dep

    # Extract the restrictions on argument-1:
    jac_restrictions = frozenset(tuple(ma.restriction for ma in extract_modified_arguments(accterm.term, argnumber=1)))
    if not jac_restrictions:
        jac_restrictions = frozenset({0})

    insn_dep = None
    for restriction in jac_restrictions:
        if accterm.argument.index:
            for i in range(world_dimension()):
                insn_dep = emit_sumfact_kernel(i, restriction, insn_dep)
        else:
            insn_dep = emit_sumfact_kernel(None, restriction, insn_dep)


def _sf_permutation_heuristic(permutations, stage):
    """Heuristic to choose a permutation

    - Stage 1: Pick the permutation where in permutations[1:] most
      elements are ordered by size
    - Stage 3: Pick the permutation where in permutations[:-1] most
      elements are ordered by size
    """
    def cost(perm, stage):
        cost = 0
        for i in range(0,len(perm)-2):
            if stage==1:
                if perm[i+1]>perm[i+2]:
                    cost += 1
            if stage==3:
                if perm[0]>perm[i+1]:
                    cost += 1
        return cost

    perm = min(permutations, key=lambda i:cost(i,stage))
    return perm


def _sf_flop_cost(a_matrices):
    """Computational cost of sumfactorization with this list of a_matrices
    """
    cost = 0;
    for l in range(len(a_matrices)):
        cost_m = 1
        cost_n = 1
        for i in range(l+1):
            cost_m *= a_matrices[i].rows
        for i in range(l,len(a_matrices)):
            cost_n *= a_matrices[i].cols
        cost += cost_m * cost_n
    return cost


def _sf_permutation_strategy(a_matrices, stage):
    """Choose permutation of a_matices list based on computational cost

    Note: If there are multiple permutations with the same cost a
    heuristic is used to pick one.
    """
    # Combine permutation and a_matrices list
    perm = [i for i, _ in enumerate(a_matrices)]
    perm_a_matrices = zip(perm, a_matrices)

    # Find cost for all possible permutations of a_matrices list
    perm_cost = []
    for permutation in itertools.permutations(perm_a_matrices):
        perm, series = zip(*permutation)
        cost = _sf_flop_cost(series)
        perm_cost.append((perm,cost))

    # Find minimal cost and all permutations with that cost
    _, costs = zip(*perm_cost)
    minimal_cost = min(costs)
    minimal_cost_permutations = [p[0] for p in perm_cost if p[1]==minimal_cost]

    # Use heuristic to pic one of the minimal cost permutations
    perm = _sf_permutation_heuristic(minimal_cost_permutations, stage)
    return perm


def _permute_forward(t, perm):
    tmp = []
    for pos in perm:
        tmp.append(t[pos])
    return tuple(tmp)


def _permute_backward(t, perm):
    tmp = [None]*len(t)
    for i, pos in enumerate(perm):
        tmp[pos] = t[i]
    return tuple(tmp)


@generator_factory(item_tags=("sumfactkernel",), context_tags=("kernel",), cache_key_generator=lambda a, b, s, **kw: (a, b, s, kw.get("restriction", 0)))
def sum_factorization_kernel(a_matrices,
                             buf,
                             stage,
                             insn_dep=frozenset({}),
                             additional_inames=frozenset({}),
                             preferred_position=None,
                             outshape=None,
                             restriction=0,
                             direct_input=None,
                             ):
    """Create a sum factorization kernel

    Sum factorization can be written as

    Y = R_{d-1} (A_{d-1} * ... * R_0 (A_0 * X)...)

    with:
    - X: Input rank d tensor of dimension n_0 x ... x n_{d-1}
    - Y: Output rank d tensor of dimension m_0 x ... x m_{d-1}
    - A_l: Values of 1D basis evaluations at quadrature points in l
      direction, matrix of dimension m_l x n_l
    - R_l: Transformation operator that permutes the underlying data
      vector of the rank d tensor in such a way that the fastest
      direction gets the slowest direction

    In the l'th step we have the following setup:
    - A_l: Matrix of dimensions m_l x n_l
    - X_l: Rank d tensor of dimensions n_l x ... x n_{d-1} x m_0 x ... x m_{l-1}
    - R_l: Transformation operator

    Looking at the indizes the following will happen:
    X --> [n_l,...,n_{d-1},m_0,...,m_{l-1}]
    A_l * X --> [m_l,n_l] * [n_l, ...] = [m_l,n_{l+1},...,n_{d-1},m_0,...,m_{l-1}]
    R_l (A_l*X) --> [n_{l+1},...,n_{d-1},m_0,...,m_{l-1}]

    So the multiplication with A_l is a reduction over one index and
    the transformation brings the next reduction index in the fastest
    position.

    Note: In the code below the transformation step is directly done
    in the reduction instruction by adapting the assignee!

    It can make sense to permute the order of directions. If you have
    a small m_l (e.g. stage 1 on faces) it is better to do direction l
    first. This can be done permuting:

    - The order of the A matrices.
    - Permuting the input tensor.
    - Permuting the output tensor (this assures that the directions of
      the output tensor are again ordered from 0 to d-1).

    Arguments:
    ----------
    a_matrices: An iterable of AMatrix instances
        The list of tensors to be applied to the input.
        Order of application is from 0 up.
    buf: A string identifying the flip flop buffer in use
        for intermediate results. The memory is expected to be
        pre-initialized with the input or you have to provide
        direct_input (FastDGGridOperator).
    insn_dep: An instruction ID that the first issued instruction
        should depend upon. All following ones will depend on each
        other.
    additional_inames: Instructions will be executed within those
        inames (needed for stage 3 in jacobians).
    preferred_position: Will be used in the dry run to order kernels
        when doing vectorization e.g. (dx u,dy u,dz u, u).
    outshape: Shape of the output.
    restriction: Restriction for faces values.
    direct_input: Global data structure containing input for
        sumfactorization (e.g. when using FastDGGridOperator).
    """
    if get_global_context_value("dry_run", False):
        return SumfactKernel(a_matrices, buf, stage, preferred_position, restriction), frozenset()

    ftags = ",".join(["f"]*len(a_matrices))
    ctags = ",".join(["c"]*len(a_matrices))
    vec_shape = ()
    if next(iter(a_matrices)).vectorized:
        ftags = ftags + ",vec"
        ctags = ctags + ",vec"
        vec_shape = (4,)

    # Measure times and count operations in c++ code
    if get_option("instrumentation_level") >= 4:
        timer_name = assembler_routine_name() + '_kernel' + '_stage{}'.format(stage)
        post_include('HP_DECLARE_TIMER({});'.format(timer_name), filetag='operatorfile')
        dump_accumulate_timer(timer_name)
        insn_dep = frozenset({instruction(code="HP_TIMER_START({});".format(timer_name),
                                          depends_on=insn_dep,
                                          within_inames=additional_inames)})

    # Put a barrier before the sumfactorization kernel
    insn_dep = frozenset({barrier(depends_on=insn_dep,
                                  within_inames=additional_inames,
                                  )})

    # Decide in which order we want to process directions in the
    # sumfactorization. A clever ordering can lead to a reduced
    # complexity. This will e.g. happen at faces where we only have
    # one quadratue point m_l=1 if l is the normal direction of the
    # face.
    #
    # Rule of thumb: small m's early and large n's late.
    perm = _sf_permutation_strategy(a_matrices, stage)

    # Permute a_matrices
    a_matrices = _permute_forward(a_matrices, perm)

    # Product of all matrices
    for l, a_matrix in enumerate(a_matrices):
        # Compute the correct shapes of in- and output matrices of this matrix-matrix multiplication
        # and get inames that realize the product.
        inp_shape = (a_matrix.cols,) + tuple(mat.cols for mat in a_matrices[l + 1:]) + tuple(mat.rows for mat in a_matrices[:l])
        out_shape = (a_matrix.rows,) + tuple(mat.cols for mat in a_matrices[l + 1:]) + tuple(mat.rows for mat in a_matrices[:l])
        out_inames = tuple(sumfact_iname(length, "out_inames_" + str(k)) for k, length in enumerate(out_shape))
        vec_iname = ()
        if a_matrix.vectorized:
            iname = sumfact_iname(4, "vec")
            vec_iname = (prim.Variable(iname),)
            transform(lp.tag_inames, [(iname, "vec")])

        # A trivial reduction is implemented as a product, otherwise we run into
        # a code generation corner case producing way too complicated code. This
        # could be fixed upstream, but the loopy code realizing reductions is not
        # trivial and the priority is kind of low.
        if a_matrix.cols != 1:
            k = sumfact_iname(a_matrix.cols, "red")
            k_expr = prim.Variable(k)
        else:
            k_expr = 0

        # Setup the input of the sum factorization kernel. In the
        # first matrix multiplication this can be taken from
        # * an input temporary (default)
        # * a global data structure (if FastDGGridOperator is in use)
        # * a value from a global data structure, broadcasted to a vector type (vectorized + FastDGGridOperator)
        input_inames = (k_expr,) + tuple(prim.Variable(j) for j in out_inames[1:])
        if l == 0 and direct_input is not None:
            # See comment bellow
            input_inames = _permute_backward(input_inames, perm)

            globalarg(direct_input, dtype=np.float64, shape=inp_shape)
            if a_matrix.vectorized:
                input_summand = prim.Call(prim.Variable("Vec4d"),
                                          (prim.Subscript(prim.Variable(direct_input),
                                                          input_inames)))
            else:
                input_summand = prim.Subscript(prim.Variable(direct_input),
                                               input_inames + vec_iname)
        else:
            # If we did permute the order of a matrices above we also
            # permuted the order of out_inames. Unfortunately the
            # order of our input is from 0 to d-1. This means we need
            # to permute _back_ to get the right coefficients.
            if l == 0:
                inp_shape = _permute_backward(inp_shape, perm)
                input_inames = _permute_backward(input_inames, perm)

            # Get a temporary that interprets the base storage of the input
            # as a column-major matrix. In later iteration of the amatrix loop
            # this reinterprets the output of the previous iteration.
            inp = get_buffer_temporary(buf,
                                       shape=inp_shape + vec_shape,
                                       dim_tags=ftags)

            # The input temporary will only be read from, so we need to silence the loopy warning
            silenced_warning('read_no_write({})'.format(inp))

            input_summand = prim.Subscript(prim.Variable(inp),
                                           input_inames + vec_iname)

        switch_base_storage(buf)

        # Get a temporary that interprets the base storage of the output.
        #
        # Note: In this step the reordering of the fastest directions
        # is happening. The new direction (out_inames[0]) and the
        # corresponding shape (out_shape[0]) goes to the end (slowest
        # direction) and everything stays column major (ftags->fortran
        # style).
        #
        # If we are in the last step we reverse the permutation.
        output_shape = tuple(out_shape[1:]) + (out_shape[0],)
        if l == len(a_matrices)-1:
            output_shape = _permute_backward(output_shape, perm)
        out = get_buffer_temporary(buf,
                                   shape=output_shape + vec_shape,
                                   dim_tags=ftags)

        # Write the matrix-matrix multiplication expression
        matprod = Product((prim.Subscript(prim.Variable(a_matrix.name),
                                          (prim.Variable(out_inames[0]), k_expr) + vec_iname),
                           input_summand))

        # ... which may be a reduction, if k>0
        if a_matrix.cols != 1:
            matprod = lp.Reduction("sum", k, matprod)

        # Here we also move the new direction (out_inames[0]) to the
        # end and reverse permutation
        output_inames = tuple(prim.Variable(i) for i in out_inames[1:]) + (prim.Variable(out_inames[0]),)
        if l == len(a_matrices)-1:
            output_inames = _permute_backward(output_inames, perm)
        assignee = prim.Subscript(prim.Variable(out), output_inames + vec_iname)

        # Issue the reduction instruction that implements the multiplication
        # at the same time store the instruction ID for the next instruction to depend on
        insn_dep = frozenset({instruction(assignee=assignee,
                                          expression=matprod,
                                          forced_iname_deps=frozenset([iname for iname in out_inames]).union(additional_inames),
                                          forced_iname_deps_is_final=True,
                                          depends_on=insn_dep,
                                          )
                              })

    # Measure times and count operations in c++ code
    if get_option("instrumentation_level") >= 4:
        insn_dep = frozenset({instruction(code="HP_TIMER_STOP({});".format(timer_name),
                                          depends_on=insn_dep,
                                          within_inames=additional_inames)})
        if stage == 1:
            qp_timer_name = assembler_routine_name() + '_kernel' + '_quadratureloop'
            post_include('HP_DECLARE_TIMER({});'.format(timer_name), filetag='operatorfile')
            dump_accumulate_timer(timer_name)
            insn_dep = instruction(code="HP_TIMER_START({});".format(qp_timer_name),
                                   depends_on=insn_dep)

    if outshape is None:
        assert stage == 3
        outshape = tuple(mat.rows for mat in a_matrices)

    dim_tags = ",".join(['f'] * len(outshape))

    if next(iter(a_matrices)).vectorized:
        outshape = outshape + vec_shape
        # This is a 'bit' hacky: In stage 3 we need to return something with vectag, in stage 1 not.
        if stage == 1:
            dim_tags = dim_tags + ",c"
        else:
            dim_tags = dim_tags + ",vec"

    out = get_buffer_temporary(buf,
                               shape=outshape,
                               dim_tags=dim_tags,
                               )
    silenced_warning('read_no_write({})'.format(out))

    return next(iter(a_matrices)).output_to_pymbolic(out), insn_dep
